export class Pair<F, S> {

    public readonly first: F;
    public readonly second: S;

    public constructor(first: F, second: S) {
        this.first = first;
        this.second = second;
    }
    
}
