/** all constants related to any input from match engine */
const INPUT_CONSTANTS = {
  DONE: 'D_DONE',
  RESEARCH_POINTS: 'rp',
  RESOURCES: 'r',
  UNITS: 'u',
  CITY: 'c',
  CITY_TILES: 'ct',
  ROADS: 'ccd',
}

module.exports = {
  INPUT_CONSTANTS,
}
